@@ -46,12 +46,15 @@ static GtkTreeModel *devices_model;
 static guint num_adapters_present = 0;
 static guint num_adapters_powered = 0;
 static gboolean show_icon_pref = TRUE;
+static gboolean discover_lock = FALSE;
 
 #define PREF_DIR		"/apps/bluetooth-manager"
 #define PREF_SHOW_ICON		PREF_DIR "/show_icon"
 
 #define KEYBOARD_PREFS		"gnome-keyboard-properties"
 #define MOUSE_PREFS		"gnome-mouse-properties"
+#define SOUND_PREFS		"gnome-volume-control"
+#define SOUND_PREFS_CMDLINE	SOUND_PREFS " -p hardware"
 
 enum {
 	CONNECTED,
@@ -101,13 +104,37 @@ select_device_changed(BluetoothChooser *sel,
 				GTK_RESPONSE_ACCEPT, address != NULL);
 }
 
+static void
+mount_finish_cb (GObject *source_object,
+		 GAsyncResult *res,
+		 gpointer user_data)
+{
+	GError *error = NULL;
+	char *uri;
+
+	if (g_file_mount_enclosing_volume_finish (G_FILE (source_object),
+						  res, &error) == FALSE) {
+		g_printerr ("Failed to mount OBEX volume: %s", error->message);
+		g_error_free (error);
+		return;
+	}
+
+	uri = g_file_get_uri (G_FILE (source_object));
+	if (gtk_show_uri (NULL, uri, GDK_CURRENT_TIME, &error) == FALSE) {
+		g_printerr ("Failed to open %s: %s", uri, error->message);
+		g_error_free (error);
+	}
+	g_free (uri);
+}
+
 void browse_callback(GObject *widget, gpointer user_data)
 {
-	char *address, *cmd;
+	GFile *file;
+	char *address, *uri;
 
 	address = g_strdup (g_object_get_data (widget, "address"));
 	if (address == NULL) {
-		GtkWidget *dialog, *selector;
+		GtkWidget *dialog, *selector, *content_area;
 		int response_id;
 
 		dialog = gtk_dialog_new_with_buttons(_("Select Device to Browse"), NULL,
@@ -120,7 +147,8 @@ void browse_callback(GObject *widget, gpointer user_data)
 		gtk_window_set_default_size(GTK_WINDOW(dialog), 480, 400);
 
 		gtk_container_set_border_width (GTK_CONTAINER (dialog), 5);
-		gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->vbox), 2);
+		content_area = gtk_dialog_get_content_area (GTK_DIALOG (dialog));
+		gtk_box_set_spacing (GTK_BOX (content_area), 2);
 
 		selector = bluetooth_chooser_new(_("Select device to browse"));
 		gtk_container_set_border_width(GTK_CONTAINER(selector), 5);
@@ -134,7 +162,7 @@ void browse_callback(GObject *widget, gpointer user_data)
 			     NULL);
 		g_signal_connect(selector, "selected-device-changed",
 				 G_CALLBACK(select_device_changed), dialog);
-		gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), selector);
+		gtk_container_add (GTK_CONTAINER (content_area), selector);
 
 		address = NULL;
 		response_id = gtk_dialog_run (GTK_DIALOG (dialog));
@@ -147,14 +175,14 @@ void browse_callback(GObject *widget, gpointer user_data)
 			return;
 	}
 
-	cmd = g_strdup_printf("%s --no-default-window \"obex://[%s]\"",
-			      "nautilus", address);
+	uri = g_strdup_printf ("obex://[%s]/", address);
 	g_free (address);
 
-	if (!g_spawn_command_line_async(cmd, NULL))
-		g_printerr("Couldn't execute command: %s\n", cmd);
+	file = g_file_new_for_uri (uri);
+	g_free (uri);
 
-	g_free (cmd);
+	g_file_mount_enclosing_volume (file, G_MOUNT_MOUNT_NONE, NULL, NULL, mount_finish_cb, NULL);
+	g_object_unref (file);
 }
 
 void sendto_callback(GObject *widget, gpointer user_data)
@@ -211,6 +239,14 @@ static void mouse_callback(GObject *widget, gpointer user_data)
 		g_printerr("Couldn't execute command: %s\n", command);
 }
 
+static void sound_callback(GObject *widget, gpointer user_data)
+{
+	const char *command = SOUND_PREFS_CMDLINE;
+
+	if (!g_spawn_command_line_async(command, NULL))
+		g_printerr("Couldn't execute command: %s\n", command);
+}
+
 void wizard_callback(GObject *widget, gpointer user_data)
 {
 	const char *command = "bluetooth-wizard";
@@ -238,7 +274,9 @@ bluetooth_discoverable_callback (GtkToggleAction *toggleaction, gpointer user_da
 	gboolean discoverable;
 
 	discoverable = gtk_toggle_action_get_active (toggleaction);
+	discover_lock = TRUE;
 	bluetooth_client_set_discoverable (client, discoverable);
+	discover_lock = FALSE;
 }
 
 static gboolean program_available(const char *program)
@@ -390,10 +428,7 @@ update_menu_items (void)
 	if (enabled == FALSE)
 		return;
 
-	object = gtk_builder_get_object (xml, "browse-device");
-	gtk_action_set_sensitive (GTK_ACTION (object),
-				  program_available ("nautilus"));
-
+	gtk_action_group_set_sensitive (GTK_ACTION_GROUP (object), TRUE);
 }
 
 static void
@@ -420,6 +455,12 @@ update_discoverability (GtkTreeIter *iter)
 	gboolean discoverable;
 	GObject *object;
 
+	/* Avoid loops from changing the UI */
+	if (discover_lock != FALSE)
+		return;
+
+	discover_lock = TRUE;
+
 	object = gtk_builder_get_object (xml, "discoverable");
 
 	if (iter == NULL) {
@@ -433,6 +474,8 @@ update_discoverability (GtkTreeIter *iter)
 
 	gtk_action_set_visible (GTK_ACTION (object), TRUE);
 	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (object), discoverable);
+
+	discover_lock = FALSE;
 }
 
 static void
@@ -547,7 +590,7 @@ action_set_bold (GtkUIManager *manager, GtkAction *action, const char *path)
 
 	label = gtk_action_get_label (action);
 	str = g_strdup_printf ("<b>%s</b>", label);
-	gtk_label_set_markup (GTK_LABEL (GTK_BIN (widget)->child), str);
+	gtk_label_set_markup (GTK_LABEL (gtk_bin_get_child (GTK_BIN (widget))), str);
 	g_free (str);
 }
 
@@ -859,6 +902,16 @@ update_device_list (GtkTreeIter *parent)
 					       menu_merge_id,
 					       G_CALLBACK (mouse_callback));
 			}
+			if ((type == BLUETOOTH_TYPE_HEADSET ||
+			     type == BLUETOOTH_TYPE_HEADPHONES ||
+			     type == BLUETOOTH_TYPE_OTHER_AUDIO) && program_available (SOUND_PREFS)) {
+				add_menu_item (address,
+					       "sound",
+					       _("Open Sound Preferences..."),
+					       uimanager,
+					       menu_merge_id,
+					       G_CALLBACK (sound_callback));
+			}
 		} else {
 			gtk_action_set_label (action, name);
 
@@ -1015,6 +1068,7 @@ int main(int argc, char *argv[])
 	if (option_debug == FALSE) {
 		app = unique_app_new ("org.gnome.Bluetooth.applet", NULL);
 		if (unique_app_is_running (app)) {
+			gdk_notify_startup_complete ();
 			g_warning ("Applet is already running, exiting");
 			return 0;
 		}
